package org.gcube.accounting.analytics.persistence.postgresql;

import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.Collection;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.gcube.accounting.analytics.Filter;
import org.gcube.accounting.analytics.TemporalConstraint;
import org.gcube.accounting.analytics.TemporalConstraint.AggregationMode;
import org.gcube.accounting.analytics.TemporalConstraint.CalendarEnum;
import org.gcube.accounting.datamodel.UsageRecord;
import org.gcube.accounting.datamodel.aggregation.AggregatedServiceUsageRecord;
import org.gcube.accounting.datamodel.aggregation.AggregatedStorageUsageRecord;
import org.gcube.accounting.utility.postgresql.PostgreSQLQuery;
import org.gcube.accounting.utility.postgresql.RecordToDBFields;
import org.gcube.accounting.utility.postgresql.RecordToDBMapping;
import org.gcube.documentstore.records.AggregatedRecord;

public class Query extends PostgreSQLQuery {
	
	public static final String DATE_OF_TIMESERIES_AS_FIELD = AggregatedRecord.START_TIME;
	
	public static final String NESTED_TABLE_NAME = "nested";
	
	private Class<? extends AggregatedRecord<?, ?>> clz;
	private final RecordToDBFields recordToDBFields;
	
	private Set<String> requestedTableField;
	
	protected TemporalConstraint temporalConstraint;
	protected List<Filter> filters;
	protected Set<String> contexts;

	private String tableFieldToRequest;
	private String orderByField;
	private Integer limit;
	
	private String recordId;
	
	protected String orConditionKey;
	protected Set<String> orConditionValues;
	
	public void setOrConditionKey(String orConditionKey) {
		this.orConditionKey = orConditionKey;
	}
	
	public void setOrConditionValues(Collection<String> orConditionValues) {
		this.orConditionValues = new HashSet<>(orConditionValues);
	}
	
	public void addOrConditionValue(String conditionValue) {
		if(this.orConditionValues == null) {
			this.orConditionValues = new HashSet<>();
		}
		this.orConditionValues.add(conditionValue);
	}
	
	public Query(Class<? extends AggregatedRecord<?, ?>> clz) throws Exception {
		this.clz = clz;
		this.recordToDBFields = RecordToDBMapping.getRecordToDBFields(clz);
	}
	
	public Set<String> getRequestedTableField() {
		return requestedTableField;
	}

	public void setTemporalConstraint(TemporalConstraint temporalConstraint) {
		this.temporalConstraint = temporalConstraint;
	}

	public void setFilters(List<Filter> filters) {
		this.filters = filters;
	}
	
	public void setContexts(Set<String> contexts) {
		this.contexts = contexts;
	}
	
	public void setTableFieldToRequest(String tableFieldToRequest) {
		this.tableFieldToRequest = tableFieldToRequest;
	}

	public void setOrderByField(String orderByField) {
		this.orderByField = orderByField;
	}

	public void setLimit(Integer limit) {
		this.limit = limit;
	}
	
	public void addContext(String context) {
		if(this.contexts == null) {
			this.contexts = new HashSet<>();
		}
		this.contexts.add(context);
	}
	
	public void setRecordId(String recordId) {
		this.recordId = recordId;
	}
	
	public RecordToDBFields getRecordToDBMapper() {
		return recordToDBFields;
	}

	protected String getTableField(String fieldName) {
		return recordToDBFields.getTableField(fieldName);
	}
	
	protected void appendTableField(String fieldName) {
		stringBuffer.append(getTableField(fieldName));
	}
	
	public static String getFormattedDate(Calendar calendar) {
		SimpleDateFormat simpleDateFormat = new SimpleDateFormat(DATETIME_PATTERN);
		String date = simpleDateFormat.format(calendar.getTime());
		return date;
	}
	
	protected void addFilters() {
		if(filters!=null && filters.size()>0) {
			// The first filter if the time_bucket
			stringBuffer.append(" AND ");
			boolean first = true;
			for(Filter filter : filters) {
				if(first) {
					first = false;
				}else {
					stringBuffer.append(" AND ");
				}
				appendTableField(filter.getKey());
				stringBuffer.append("=");
				appendValue(filter.getValue());
			}
		}
	}
	
	protected void addOrConditions(String key, Set<String> values) {
		if(values!=null && values.size()>0) {
			// The first filter if the time_bucket
			stringBuffer.append(" AND (");
			boolean first = true;
			for(String value : values) {
				if(first) {
					first = false;
				}else {
					stringBuffer.append(" OR ");
				}
				appendTableField(key);
				stringBuffer.append("=");
				appendValue(value);
			}
			stringBuffer.append(")");
		}
	}
	
	
	protected void addContextFilter() {
		addOrConditions(UsageRecord.SCOPE, contexts);
	}
	
	protected void addOrConditionFilter() {
		addOrConditions(orConditionKey, orConditionValues);
	}
	
	protected void addEmittedFields(boolean addNested) throws Exception {
		Set<String> aggregatedField = clz.newInstance().getAggregatedFields();
		for(String fieldName : aggregatedField) {
			String dbField = getTableField(fieldName);
			switch (fieldName) {
				case AggregatedRecord.START_TIME: 
				case AggregatedRecord.END_TIME:
				case AggregatedRecord.AGGREGATED:
					continue;

				case AggregatedRecord.OPERATION_COUNT:
					stringBuffer.append(", SUM(");
					if(addNested) {
						stringBuffer.append(NESTED_TABLE_NAME);
						stringBuffer.append(".");
					}
					stringBuffer.append(dbField);
					stringBuffer.append(") AS ");
					break;
				
				case AggregatedStorageUsageRecord.DATA_VOLUME:
					stringBuffer.append(", SUM(");
					if(addNested) {
						stringBuffer.append(NESTED_TABLE_NAME);
						stringBuffer.append(".");
					}
					stringBuffer.append(dbField);
					stringBuffer.append(") AS ");
					break;
					
				// WEIGHTED AVERAGE
				case AggregatedServiceUsageRecord.DURATION:
					stringBuffer.append(", ROUND(SUM(");
					if(addNested) {
						stringBuffer.append(NESTED_TABLE_NAME);
						stringBuffer.append(".");
					}
					stringBuffer.append(dbField);
					stringBuffer.append("*");
					if(addNested) {
						stringBuffer.append(NESTED_TABLE_NAME);
						stringBuffer.append(".");
					}
					appendTableField(AggregatedRecord.OPERATION_COUNT);
					stringBuffer.append(")/SUM(");
					if(addNested) {
						stringBuffer.append(NESTED_TABLE_NAME);
						stringBuffer.append(".");
					}
					appendTableField(AggregatedRecord.OPERATION_COUNT);
					stringBuffer.append(")) AS ");
					break;
					
				case AggregatedServiceUsageRecord.MAX_INVOCATION_TIME:
					stringBuffer.append(", MAX(");
					if(addNested) {
						stringBuffer.append(NESTED_TABLE_NAME);
						stringBuffer.append(".");
					}
					stringBuffer.append(dbField);
					stringBuffer.append(") AS ");
					break;
				
				case AggregatedServiceUsageRecord.MIN_INVOCATION_TIME:
					stringBuffer.append(", MIN(");
					if(addNested) {
						stringBuffer.append(NESTED_TABLE_NAME);
						stringBuffer.append(".");
					}
					stringBuffer.append(dbField);
					stringBuffer.append(") AS ");
					break;
					
				default:
					stringBuffer.append(", ");
					break;
			}
			
			stringBuffer.append(dbField);
			requestedTableField.add(dbField);
		}
	}
	
	protected String getTimeBucketCalendarInterval(AggregationMode aggregationMode) {
		CalendarEnum calendarEnum = CalendarEnum.values()[aggregationMode.ordinal()];
		return calendarEnum.name().toLowerCase();
	}
	
	private void newQuery() {
		stringBuffer = new StringBuffer();
		requestedTableField = new HashSet<>();
		stringBuffer.append("SELECT ");
	}
	
	protected void addTemporalConstraintToQuery() {
		stringBuffer.append(" WHERE ");
		String tableField = getTableField(AggregatedRecord.START_TIME);
		stringBuffer.append(tableField);
		stringBuffer.append(" > ");
		appendValue(temporalConstraint.getAlignedStartTime());
		
		stringBuffer.append(" AND ");
		tableField = getTableField(AggregatedRecord.END_TIME);
		stringBuffer.append(tableField);
		stringBuffer.append(" < ");
		appendValue(temporalConstraint.getAlignedEndTime());
	}
	
	protected void addDateGroupBy(boolean addNested) {
		stringBuffer.append(" GROUP BY ");
		if(addNested) {
			stringBuffer.append(NESTED_TABLE_NAME);
			stringBuffer.append(".");
		}
		appendTableField(DATE_OF_TIMESERIES_AS_FIELD);
	}
	
	protected void addDateOrderBy(boolean addNested) {
		stringBuffer.append(" ORDER BY ");
		if(addNested) {
			stringBuffer.append(NESTED_TABLE_NAME);
			stringBuffer.append(".");
		}
		appendTableField(DATE_OF_TIMESERIES_AS_FIELD);
		stringBuffer.append(" ASC");
	}
	
	protected void addGroupAndOrderByForOrderByField() {
		stringBuffer.append(" GROUP BY ");
		String dbField = getTableField(tableFieldToRequest);
		stringBuffer.append(dbField);
		stringBuffer.append(" ORDER BY ");
		dbField = getTableField(orderByField);
		stringBuffer.append(dbField);
		stringBuffer.append(" DESC");
	}
	
	protected void addLimit() {
		stringBuffer.append(" LIMIT ");
		if(limit != null) {
			stringBuffer.append(limit.toString());
		}else {
			stringBuffer.append(50);
		}
	}

	protected void addRequestedField() {
		String dbField = getTableField(tableFieldToRequest);
		stringBuffer.append(dbField);
		requestedTableField.add(dbField);
		stringBuffer.append(", SUM(");
		dbField = getTableField(orderByField);
		stringBuffer.append(dbField);
		requestedTableField.add(dbField);
		stringBuffer.append(") AS ");
		stringBuffer.append(dbField);
	}
	
	protected void addRequestedDate() {
		String calendarInterval = getTimeBucketCalendarInterval(temporalConstraint.getAggregationMode());
		String tableField = getTableField(DATE_OF_TIMESERIES_AS_FIELD);
		stringBuffer.append("date_trunc('");
		stringBuffer.append(calendarInterval);
		stringBuffer.append("', ");
		stringBuffer.append(tableField);
		stringBuffer.append(" AT TIME ZONE 'UTC') ");
		stringBuffer.append(tableField);
		requestedTableField.add(tableField);
	}
	
	public String getTimeSeriesQuery() throws Exception {
		newQuery();
		
		stringBuffer.append(NESTED_TABLE_NAME);
		stringBuffer.append(".");
		String tableField = getTableField(DATE_OF_TIMESERIES_AS_FIELD);
		stringBuffer.append(tableField);
		
		addEmittedFields(true);
		
		/* Nested Query */
		stringBuffer.append(" FROM ( SELECT ");
		
		addRequestedDate();
		addEmittedFields(false);
		
		stringBuffer.append(" FROM ");
		stringBuffer.append(recordToDBFields.getTableName());
		
		addTemporalConstraintToQuery();
		
		addFilters();
		addContextFilter();
		
		addDateGroupBy(false);
		addDateOrderBy(false);
		
		stringBuffer.append(") AS ");
		stringBuffer.append(NESTED_TABLE_NAME);
		
		/* Nested Query */
		
		addDateGroupBy(true);
		addDateOrderBy(true);
		
		return stringBuffer.toString();
	}
	
	public String getNextPossibleValueQuery() throws Exception {
		newQuery();
		
		addRequestedField();
		
		stringBuffer.append(" FROM ");
		stringBuffer.append(recordToDBFields.getTableName());
		
		addTemporalConstraintToQuery();
		
		addFilters();
		addContextFilter();
		
		addGroupAndOrderByForOrderByField();
		addLimit();
		
		return stringBuffer.toString();
	}

	public String getRecordQuery(){
		newQuery();
		stringBuffer.append("* ");
		stringBuffer.append(" FROM ");
		stringBuffer.append(recordToDBFields.getTableName());
		stringBuffer.append(" WHERE ");
		stringBuffer.append(" id=");
		appendString(recordId);
		return stringBuffer.toString();
	}

	public String getDinstinctValuesQuery() {
		newQuery();
		
		String dbField = getTableField(tableFieldToRequest);
		requestedTableField.add(dbField);
		
		stringBuffer.append("DISTINCT ");
		stringBuffer.append(dbField);
		stringBuffer.append(" FROM ");
		stringBuffer.append(recordToDBFields.getTableName());
		
		return stringBuffer.toString();
	}
	
}
